// ------------------------------------------------------------------------
// audioio-cdr.cpp: CDDA/CDR audio file format input/output
// Copyright (C) 1999 Kai Vehmanen (kaiv@wakkanet.fi)
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA
// ------------------------------------------------------------------------

#include <cmath>
#include <string>
#include <cstring>
#include <cassert>

#include "samplebuffer.h"
#include "audioio.h"
#include "audioio-cdr.h"

#include "eca-error.h"
#include "eca-debug.h"

CDRFILE::CDRFILE(const string& name, const SIMODE mode, const
		 AIO_PARAMS& fmt, int bsize) 
  :  AUDIO_IO_DEVICE(name, mode, fmt, bsize) {

  switch(mode) {
  case si_read:
    {
      fobject=fopen(name.c_str(),"rb");
      if (!fobject)
	throw(new ECA_ERROR("AUDIOIO-CDR", "Can't open " + name + " for reading."));
      set_length_in_bytes();
      break;
    }
  case si_write: 
    {
      fobject=fopen(name.c_str(),"wb");
      if (!fobject) 
	throw(new ECA_ERROR("AUDIOIO-CDR","Can't open " + name + " for writing."));
      break;
    }
  case si_readwrite:
    {
      fobject=fopen(name.c_str(),"r+b");
      if (!fobject) {
	fobject=fopen(name.c_str(),"w+b");
	if (!fobject)
	  throw(new ECA_ERROR("AUDIOIO-CDR","Can't open " + name + " for read-wre."));
      }
      set_length_in_bytes();
      break;
    }
  }

  AIO_PARAMS sffmt;
  
  sffmt.channels = 2;
  sffmt.bits = 16;
  sffmt.srate = 44100;
  
  format(sffmt);
  position_in_samples(0, true);
}

CDRFILE::~CDRFILE(void) {
  pad_to_sectorsize();

  fclose(fobject);
}

void CDRFILE::get_sample(SAMPLE_BUFFER* t) {
  t->reserve_buffer(buffersize());
  samples_read = fread(t->iobuf_uchar, format().align, buffersize(), fobject);
  
  if (!ferror(fobject)) {  
    t->iobuf_to_buffer(samples_read, true, format().bits, format().channels, format().srate);
    position_in_samples(position_in_samples() + t->length_in_samples());
    if (feof(fobject)) finished(true);
  }
  else
    throw(new ECA_ERROR("AUDIOIO-CDR","Read error!"));
}

void CDRFILE::pad_to_sectorsize(void) {
  int padsamps = CDRFILE::sectorsize - (length_in_samples() % CDRFILE::sectorsize);

  if (padsamps == CDRFILE::sectorsize) {
    return;
  }
  for(int n = 0; n < padsamps; n++) fputc(0,fobject);
  
  assert(ftell(fobject) %  CDRFILE::sectorsize == 0);
}

void CDRFILE::put_sample(SAMPLE_BUFFER* t) {
  t->buffer_to_iobuf(false, format().bits, format().channels, format().srate);
  position_in_samples(position_in_samples() + t->length_in_samples());
  length_in_samples(position_in_samples());

  fwrite(t->iobuf_uchar, format().align, t->length_in_samples(), fobject);

  if (ferror(fobject) || t->length_in_samples() <  t->length_in_samples()) finished(true);
}

void CDRFILE::set_length_in_bytes(void) {
  long int save = ftell(fobject);
  fseek(fobject,0,SEEK_END);
  length_in_samples(ftell(fobject) / format().align);
  fseek(fobject,save,SEEK_SET);
}

void CDRFILE::position_in_samples_hook(void) {
  finished(false);
  fseek(fobject, position_in_samples() * format().align, SEEK_SET);
}
